#include "common.h"
#include "typed_value.h"

#include <stdio.h>
#include <string.h>

vtd_value_type type_of_value (TV_t *v) {
  switch (v->tv_tag) {
  case TV_Sbool:		return vtd_scalar_type (VTD_Bool_type);
  case TV_Sbool_couple:		return vtd_scalar_type (VTD_Couple_Bool_type);
  case TV_Sbool_triple:		return vtd_scalar_type (VTD_Triple_Bool_type);
  case TV_Sstring:		return vtd_scalar_type (VTD_String_type);
  case TV_Sint:			return vtd_scalar_type (VTD_Int_type);
  case TV_Sint_couple:		return vtd_scalar_type (VTD_Couple_Int_type);
  case TV_Sint_triple:		return vtd_scalar_type (VTD_Triple_Int_type);
  case TV_Snative_int:		return vtd_scalar_type (VTD_Native_Int_type);
  case TV_Snative_int_couple:	return vtd_scalar_type (VTD_Couple_Native_Int_type);
  case TV_Snative_int_triple:	return vtd_scalar_type (VTD_Triple_Native_Int_type);
  case TV_Sint32:		return vtd_scalar_type (VTD_Int32_type);
  case TV_Sint32_couple:	return vtd_scalar_type (VTD_Couple_Int32_type);
  case TV_Sint32_triple:	return vtd_scalar_type (VTD_Triple_Int32_type);
  case TV_Sint64:		return vtd_scalar_type (VTD_Int64_type);
  case TV_Sint64_couple:	return vtd_scalar_type (VTD_Couple_Int64_type);
  case TV_Sint64_triple:	return vtd_scalar_type (VTD_Triple_Int64_type);
  case TV_Sfloat:		return vtd_scalar_type (VTD_Float_type);
  case TV_Sfloat_couple:	return vtd_scalar_type (VTD_Couple_Float_type);
  case TV_Sfloat_triple:	return vtd_scalar_type (VTD_Triple_Float_type);
  case TV_Vbool:		return vtd_vector_type (VTD_Bool_type);
  case TV_Vbool_couple:		return vtd_vector_type (VTD_Couple_Bool_type);
  case TV_Vbool_triple:		return vtd_vector_type (VTD_Triple_Bool_type);
  case TV_Vstring:		return vtd_vector_type (VTD_String_type);
  case TV_Vint:			return vtd_vector_type (VTD_Int_type);
  case TV_Vint_couple:		return vtd_vector_type (VTD_Couple_Int_type);
  case TV_Vint_triple:		return vtd_vector_type (VTD_Triple_Int_type);
  case TV_Vnative_int:		return vtd_vector_type (VTD_Native_Int_type);
  case TV_Vnative_int_couple:	return vtd_vector_type (VTD_Couple_Native_Int_type);
  case TV_Vnative_int_triple:	return vtd_vector_type (VTD_Triple_Native_Int_type);
  case TV_Vint32:		return vtd_vector_type (VTD_Int32_type);
  case TV_Vint32_couple:	return vtd_vector_type (VTD_Couple_Int32_type);
  case TV_Vint32_triple:	return vtd_vector_type (VTD_Triple_Int32_type);
  case TV_Vint64:		return vtd_vector_type (VTD_Int64_type);
  case TV_Vint64_couple:	return vtd_vector_type (VTD_Couple_Int64_type);
  case TV_Vint64_triple:	return vtd_vector_type (VTD_Triple_Int64_type);
  case TV_Vfloat:		return vtd_vector_type (VTD_Float_type);
  case TV_Vfloat_couple:	return vtd_vector_type (VTD_Couple_Float_type);
  case TV_Vfloat_triple:	return vtd_vector_type (VTD_Triple_Float_type);
  case TV_Mc_float:			/* Falls through */
  case TV_Mfortran_float:	return vtd_matrix_type (VTD_Float_type);
  default:
    assert (0);
  }
}

/* exception Wrong_communication of string */
/* ( should only be ?) catched by read_communication.c */
jmp_buf *Wrong_communication = NULL;
char *Wrong_communication_msg = NULL;

static const char *to_string_msg (TV_tag tag) {
  switch (tag) {
  case TV_Sbool:		return "bool";
  case TV_Sbool_couple:		return "bool_couple";
  case TV_Sbool_triple:		return "bool_triple";
  case TV_Sstring:		return "string";
  case TV_Sint:			return "int";
  case TV_Sint_couple:		return "int_couple";
  case TV_Sint_triple:		return "int_triple";
  case TV_Snative_int:		return "native_int";
  case TV_Snative_int_couple:	return "native_int_couple";
  case TV_Snative_int_triple:	return "native_int_triple";
  case TV_Sint32:		return "int32";
  case TV_Sint32_couple:	return "int32_couple";
  case TV_Sint32_triple:	return "int32_triple";
  case TV_Sint64:		return "int64";
  case TV_Sint64_couple:	return "int64_couple";
  case TV_Sint64_triple:	return "int64_triple";
  case TV_Sfloat:		return "float";
  case TV_Sfloat_couple:	return "float_couple";
  case TV_Sfloat_triple:	return "float_triple";
  case TV_Vbool:		return "bool_vector";
  case TV_Vbool_couple:		return "bool_couple_vector";
  case TV_Vbool_triple:		return "bool_triple_vector";
  case TV_Vstring:		return "string_vector";
  case TV_Vint:			return "int_vector";
  case TV_Vint_couple:		return "int_couple_vector";
  case TV_Vint_triple:		return "int_triple_vector";
  case TV_Vnative_int:		return "native_int_vector";
  case TV_Vnative_int_couple:	return "native_int_couple_vector";
  case TV_Vnative_int_triple:	return "native_int_triple_vector";
  case TV_Vint32:		return "int32_vector";
  case TV_Vint32_couple:	return "int32_couple_vector";
  case TV_Vint32_triple:	return "int32_triple_vector";
  case TV_Vint64:		return "int64_vector";
  case TV_Vint64_couple:	return "int64_couple_vector";
  case TV_Vint64_triple:	return "int64_triple_vector";
  case TV_Vfloat:		return "float_vector";
  case TV_Vfloat_couple:	return "float_couple_vector";
  case TV_Vfloat_triple:	return "float_triple_vector";
  case TV_Mc_float:		return "c_float_matrix";
  case TV_Mfortran_float:	return "fortran_float_matrix";
  default:
    assert (0);
  }
}

void raise_wrong_communication (char *msg) {
  if (Wrong_communication == NULL)
    assert (0);
  Wrong_communication_msg = msg;
  longjmp (*Wrong_communication, 1);
}

static void TV_raise_wrong_communication (char *s, TV_tag tag) __attribute__ ((noreturn));
static void TV_raise_wrong_communication (char *s, TV_tag tag) {
  snprintf (failbuf, sizeof (failbuf), "waiting for a %s, received a %s",
	    s, to_string_msg (tag));
  raise_wrong_communication (failbuf);
}

#define TV_TO(scalarity, what, vectorp, rettype, fieldnam) \
rettype to_##what##vectorp (TV_t *typed_value) { \
  switch (typed_value->tv_tag) { \
  case TV_##scalarity##what: \
    return typed_value->v.fieldnam; \
  default: \
    TV_raise_wrong_communication (#what #vectorp, typed_value->tv_tag); \
  } \
}

TV_TO (S, bool, , bool, b)
TV_TO (S, bool_couple, , struct cbool, cb)
TV_TO (S, bool_triple, , struct tbool, tb)
TV_TO (S, string, , struct string, str)
TV_TO (S, int, , int, i)
TV_TO (S, int_couple, , struct cint, ci)
TV_TO (S, int_triple, , struct tint, ti)
TV_TO (S, native_int, , natint, nat)
TV_TO (S, native_int_couple, , struct cnat, cnat)
TV_TO (S, native_int_triple, , struct tnat, tnat)
TV_TO (S, int32, , int32, i32)
TV_TO (S, int32_couple, , struct cint32, ci32)
TV_TO (S, int32_triple, , struct tint32, ti32)
TV_TO (S, int64, , int64, i64)
TV_TO (S, int64_couple, , struct cint64, ci64)
TV_TO (S, int64_triple, , struct tint64, ti64)
TV_TO (S, float, , double, f)
TV_TO (S, float_couple, , struct cfloat, cf)
TV_TO (S, float_triple, , struct tfloat, tf)

TV_TO (V, bool, _vector, bool *, vec.p.vb)
TV_TO (V, bool_couple, _vector, struct cbool *, vec.p.vcb)
TV_TO (V, bool_triple, _vector, struct tbool *, vec.p.vtb)
TV_TO (V, string, _vector, struct string *, vec.p.vs)
TV_TO (V, int, _vector, int *, vec.p.vi)
TV_TO (V, int_couple, _vector, struct cint *, vec.p.vci)
TV_TO (V, int_triple, _vector, struct tint *, vec.p.vti)
TV_TO (V, native_int, _vector, natint *, vec.p.vnat)
TV_TO (V, native_int_couple, _vector, struct cnat *, vec.p.vcnat)
TV_TO (V, native_int_triple, _vector, struct tnat *, vec.p.vtnat)
TV_TO (V, int32, _vector, int32 *, vec.p.vi32)
TV_TO (V, int32_couple, _vector, struct cint32 *, vec.p.vci32)
TV_TO (V, int32_triple, _vector, struct tint32 *, vec.p.vti32)
TV_TO (V, int64, _vector, int64 *, vec.p.vi64)
TV_TO (V, int64_couple, _vector, struct cint64 *, vec.p.vci64)
TV_TO (V, int64_triple, _vector, struct tint64 *, vec.p.vti64)
TV_TO (V, float, _vector, double *, vec.p.vf)
TV_TO (V, float_couple, _vector, struct cfloat *, vec.p.vcf)
TV_TO (V, float_triple, _vector, struct tfloat *, vec.p.vtf)

TV_TO (M, c_float, _matrix, double *, mat.mf)
TV_TO (M, fortran_float, _matrix, double *, mat.mf)

static TV_t *make_typed_value (TV_tag tag) {
  TV_t *tv = malloc (sizeof (TV_t));	/* XXX voir  allouer par paquets, mais free ? */
  if (tv == NULL)
    failwith ("make_typed_value: out of memory");

  memset ((void *)tv, 0, sizeof (*tv));	/* realloc doesn't clear memory */
  tv->tv_tag = tag;
  return tv;
}

/* By value version */
#define TV_of_V(kind, type, fieldnam)		\
  TV_t *TV_of_##kind (type val) {		\
  TV_t *tv = make_typed_value (TV_S##kind);	\
  tv->v.fieldnam = val;				\
  return tv;					\
}

/* By reference version */
#define TV_of_R(kind, type, fieldnam)		\
  TV_t *TV_of_##kind (type *val) {		\
  TV_t *tv = make_typed_value (TV_S##kind);	\
  tv->v.fieldnam = *val;			\
  return tv;					\
}

TV_of_V (bool, bool, b)
TV_of_R (bool_couple, struct cbool, cb)
TV_of_R (bool_triple, struct tbool, tb)
TV_of_R (string, struct string, str)
TV_of_V (int, int, i)
TV_of_R (int_couple, struct cint, ci)
TV_of_R (int_triple, struct tint, ti)
TV_of_V (native_int, natint, nat)
TV_of_R (native_int_couple, struct cnat, cnat)
TV_of_R (native_int_triple, struct tnat, tnat)
TV_of_V (int32, int32, i32)
TV_of_R (int32_couple, struct cint32, ci32)
TV_of_R (int32_triple, struct tint32, ti32)
TV_of_V (int64, int64, i64)
TV_of_R (int64_couple, struct cint64, ci64)
TV_of_R (int64_triple, struct tint64, ti64)
TV_of_V (float, double, f)
TV_of_R (float_couple, struct cfloat, cf)
TV_of_R (float_triple, struct tfloat, tf)

#define TV_of_vec(kind, type)					\
  TV_t *TV_of_##kind##_vector (struct vector v) {		\
    TV_t *tv = make_typed_value (TV_V##kind);			\
    tv->v.vec = v;						\
    return tv;							\
}

TV_of_vec (bool, bool)
TV_of_vec (bool_couple, struct cbool)
TV_of_vec (bool_triple, struct tbool)
TV_of_vec (string, struct string)
TV_of_vec (int, int)
TV_of_vec (int_couple, struct cint)
TV_of_vec (int_triple, struct tint)
TV_of_vec (native_int, natint)
TV_of_vec (native_int_couple, struct cnat)
TV_of_vec (native_int_triple, struct tnat)
TV_of_vec (int32, int32)
TV_of_vec (int32_couple, struct cint32)
TV_of_vec (int32_triple, struct tint32)
TV_of_vec (int64, int64)
TV_of_vec (int64_couple, struct cint64)
TV_of_vec (int64_triple, struct tint64)
TV_of_vec (float, double)
TV_of_vec (float_couple, struct cfloat)
TV_of_vec (float_triple, struct tfloat)

/* Convenience funcs for the mundane C programmer */
TV_t *TV_of_chars (char *s) {
  struct string str;

  str.ssize = strlen (s);
  str.str = s;
  return TV_of_string (&str);  /* str is copied in the resulting TV_of_string */
}

/* XXXX Tester avec valgrind */
TV_t *TV_of_chars_array (char **s, size_t asiz) {
  struct string *vstr = calloc (asiz, sizeof (*vstr));
  size_t i = 0;

  for (i = 0; i < asiz; i++, s++) {
    vstr[i].ssize = strlen (*s);
    vstr[i].str = *s;
  }
  return TV_of_string_vector ((struct vector){ .vsize = asiz, .p.vs = vstr });
}

TV_t *TV_of_c_float_matrix (double *m, size_t sizei, size_t sizej) {
  TV_t *tv = make_typed_value (TV_Mc_float);
  tv->v.mat.msize1 = sizei;
  tv->v.mat.msize2 = sizej;
  tv->v.mat.mf = m;
  return tv;
}

TV_t *TV_of_fortran_float_matrix (double *m, size_t sizei, size_t sizej) {
  TV_t *tv = make_typed_value (TV_Mfortran_float);
  tv->v.mat.msize1 = sizei;
  tv->v.mat.msize2 = sizej;
  tv->v.mat.mf = m;
  return tv;
}
