CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                     C
C                            Pio (Fortran)                            C
C                                                                     C
C                    Polyglot Input/Output Library                    C
C                                                                     C
C                  projet Estime, INRIA-Rocquencourt                  C
C                                                                     C
C  Copyright 2004-2011 INRIA.                                         C
C  Institut National de Recherche en Informatique et en Automatique.  C
C  All rights reserved. This file is distributed only by permission.  C
C                                                                     C
C  Francois Clement <Francois.Clement@inria.fr>                       C
C  Pierre Weis <Pierre.Weis@inria.fr>                                 C
C  Hend Ben Ameur <hbenameur@yahoo.ca>                                C
C                                                                     C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

c $Id: casting.F,v 1.4 2011-03-09 12:50:15 fclement Exp $

C     
C     Library to cast binary integers and binary floats to/from strings.
C     
      
#include "pio_constants.h"
      
C     
C     In the routines, we need characters: in Fortran, characters are
C     represented as 1-character long strings.
C     
      
      integer function char_to_int(c)
C     Casting a char into an integer.
      
      implicit none
      
C     Input argument.
      character c
      
C     Local variables
      character sj*INT_SIZE
      integer j
      
C     External function.
      logical pio_get_debug_mode
      
C     Equivalence-ing between an INT_SIZE-character long string and an
C     integer (of size 8*INT_SIZE-bit).
      equivalence (sj, j)
      
C     Masking the result to obtain only two hexadecimal digits
      j = 0
C     For little endian processors
      sj(1:1) = c
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In char_to_int:',
     +        ' c = ''', c, ''', j = ', j
      endif
      
      char_to_int = j
      
      return
      end
      
      
      character function int_to_char(i)
C     Casting an integer into a char.
      
      implicit none
      
C     Input argument.
      integer i
      
C     Local variables.
      character c, sj*INT_SIZE
      integer j
      
C     External function.
      logical pio_get_debug_mode
      
C     Equivalence-ing between an INT_SIZE-character long string and an
C     integer (of size 8*INT_SIZE-bit).
      equivalence (sj, j)
      
C     Masking the argument to obtain only two hexadecimal digits
      j = mod(i, 256)
C     For little endian processors
      c = sj(1:1)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In int_to_char:',
     +        ' j = ', j, ', c = ''', c, ''''
      endif
      
      int_to_char = c
      
      return
      end
      
      
      integer function string_to_int_bin(s)
C     Casting an INT_SIZE-character long string into an integer
C     (of size 8*INT_SIZE-bit).
      
      implicit none
      
C     Input argument.
      character s*INT_SIZE
      
C     Local variables.
      character sj*INT_SIZE
      integer j
      
C     Equivalence-ing between an INT_SIZE-character long string and an
C     integer (of size 8*INT_SIZE-bit).
      equivalence (sj, j)
      
      sj = s
      
      string_to_int_bin = j
      
      return
      end
      
      
      character*INT_SIZE function int_to_string_bin(i)
C     Casting an integer (of size 8*INT_SIZE-bit) into an
C     INT_SIZE-character long string.
      
      implicit none
      
C     Input argument.
      integer i
      
C     Local variables.
      character sj*INT_SIZE
      integer j
      
C     Equivalence-ing between an INT_SIZE-character long string and an
C     integer (of size 8*INT_SIZE-bit).
      equivalence (sj, j)
      
      j = i
      
      int_to_string_bin = sj
      
      return
      end
      
      
      FLOAT function string_to_float_bin(s)
C     Casting a FLOAT_SIZE-character long string into an
C     8*FLOAT_SIZE-bit float.
      
      implicit none
      
C     Input argument.
      character s*FLOAT_SIZE
      
C     Local variables.
      character sg*FLOAT_SIZE
      FLOAT g
      
C     Equivalence-ing between a FLOAT_SIZE-character long string and an
C     8FLOAT_SIZE-bit float.
      equivalence (sg, g)
      
      sg = s
      
      string_to_float_bin = g
      
      return
      end
      
      
      character*FLOAT_SIZE function float_to_string_bin(f)
C     Casting an 8*FLOAT_SIZE-bit float into a FLOAT_SIZE-character
C     long string.
      
      implicit none
      
C     Input argument.
      FLOAT f
      
C     Local variables.
      character sg*FLOAT_SIZE
      FLOAT g
      
C     Equivalence-ing between an FLOAT_SIZE-character long string and a
C     8*FLOAT_SIZE-bit float
      equivalence (sg, g)
      
      g = f
      
      float_to_string_bin = sg
      
      return
      end
