CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                     C
C                            Pio (Fortran)                            C
C                                                                     C
C                    Polyglot Input/Output Library                    C
C                                                                     C
C                  projet Estime, INRIA-Rocquencourt                  C
C                                                                     C
C  Copyright 2004-2011 INRIA.                                         C
C  Institut National de Recherche en Informatique et en Automatique.  C
C  All rights reserved. This file is distributed only by permission.  C
C                                                                     C
C  Francois Clement <Francois.Clement@inria.fr>                       C
C  Pierre Weis <Pierre.Weis@inria.fr>                                 C
C  Hend Ben Ameur <hbenameur@yahoo.ca>                                C
C                                                                     C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

      
C $Id: print_typed_value.F,v 1.5 2011-03-10 10:41:56 fclement Exp $
      
C     
C     Routines to print a typed value into the printing buffer.
C     
      
#include "pio_constants.h"
      
      subroutine buffer_print_size(sbr, sz)
C     Print a Size in the sens of the PIO library.
      
      implicit none
      
      include 'buffers_definition.fh'
      
C     Input arguments.
      character sbr*(*)
      integer sz
      
C     External function.
      logical pio_get_debug_mode
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr, ': size = ', sz
      endif
      
      call buffer_print_char(sbr, '<')
      call buffer_print_natural(sbr, sz)
      call buffer_print_char(sbr, '>')
      
      return
      end
      
      
      subroutine print_string_value(i, j, k, dim1, l, tup)
C     Print a string value.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, l, tup
      
C     Local variables.
      character sbr*18, sv(MAX_STRING_ITEMS)*MAX_STRING_LENGTH, c
      integer nl, lenv(MAX_STRING_ITEMS), m
      
C     External functions.
      logical pio_get_debug_mode
      character int_to_char
      
      sbr = 'print_string_value'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', l = ', l
      endif
      
      if (dim1 .ne. 1 .and. tup .ne. 1) then
         call error_assertion_failure
     +        (sbr, 'vectors of tuple of strings are forbidden.')
      endif
      if (k .ne. 1 .and. l .ne. 1) then
         call error_assertion_failure
     +        (sbr, 'vectors of tuple of strings are forbidden.')
      endif
      if (j .ne. 1) then
         call error_assertion_failure
     +        (sbr, 'matrices of strings are forbidden.')
      endif

      call fetch_value_lengths(sbr, PIO_OUT, i, nl, lenv)
      call rand_fetch_string_compound_value
     +     (sbr, PIO_OUT, i, nl, lenv, sv)
      
C     Newline character '\n'.
      c = int_to_char(10)
      
C     m = k when l=1, m = l when k=1.
      m = k*l
      
      call buffer_print_size(sbr, lenv(m))
      call buffer_print_char(sbr, c)
      call buffer_print_delimited_string(sbr, sv(m), lenv(m))
      
      return
      end
      
      
      subroutine print_int_value(i, j, k, dim1, l, tup)
C     Print an integer value.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, l, tup
      
C     Local variables.
      character sbr*15
      integer idatum
      
C     External functions.
      logical pio_get_debug_mode
      integer rand_fetch_int_value
      
      sbr = 'print_int_value'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', l = ', l
      endif
      
      idatum =
     +     rand_fetch_int_value(sbr, PIO_OUT, i, j, k, dim1, l, tup)
      
      call buffer_print_int(sbr, idatum)
      
      return
      end
      
      
      subroutine print_float_value(i, j, k, dim1, l, tup)
C     Print a float value.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, l, tup
      
C     Local variables.
      character sbr*17
      FLOAT fdatum
      
C     External functions.
      logical pio_get_debug_mode, communication_mode_is_binary
      FLOAT rand_fetch_float_value
      
      sbr = 'print_float_value'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', l = ', l
      endif
      
      fdatum =
     +     rand_fetch_float_value(sbr, PIO_OUT, i, j, k, dim1, l, tup)
      
      if (communication_mode_is_binary()) then
         call buffer_print_binary_float(sbr, fdatum)
      else
         call buffer_print_ascii_float(sbr, fdatum)
      endif
      
      return
      end
      
      
      subroutine print_simple(i, j, k, dim1, l, tup, typ)
C     Print a simple value.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, l, tup, typ
      
C     Local variable.
      character sbr*12
      
C     External function.
      logical pio_get_debug_mode
      
      sbr = 'print_simple'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', l = ', l,
     +        ', typ = ', typ
      endif
      
      if (typ .eq. SSTRING) then
         
         call print_string_value(i, j, k, dim1, l, tup)
         
      else if (typ .eq. SINT) then
         
         call print_int_value(i, j, k, dim1, l, tup)
         
      else if (typ .eq. SFLOAT) then
         
         call print_float_value(i, j, k, dim1, l, tup)
         
      else
         
C     impossible, already checked in print_scalar
         write(PIO_STDERR, *) 'typ = ', typ
         call error_internal_error(sbr)
         
      endif
      
      return
      end
      
      
      subroutine print_couple(i, j, k, dim1, typ)
C     Print a couple value of type typ.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, typ
      
C     Local variables.
      character sbr*12
      integer item_typ
      
C     External function.
      logical pio_get_debug_mode
      
      sbr = 'print_couple'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', typ = ', typ
      endif
      
      item_typ = MAKE_COUPLE_ITEM_TYPE(typ)
      
      call print_simple(i, j, k, dim1, 1, 2, item_typ)
      call buffer_print_string(sbr, 2, ', ')
      call print_simple(i, j, k, dim1, 2, 2, item_typ)
      
      return
      end
      
      
      subroutine print_triple(i, j, k, dim1, typ)
C     Print a triple value of type typ.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, typ
      
C     Local variables.
      character sbr*12
      integer item_typ
      
C     External function.
      logical pio_get_debug_mode
      
      sbr = 'print_triple'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', typ = ', typ
      endif
      
      item_typ = MAKE_TRIPLE_ITEM_TYPE(typ)
      
      call print_simple(i, j, k, dim1, 1, 3, item_typ)
      call buffer_print_string(sbr, 2, ', ')
      call print_simple(i, j, k, dim1, 2, 3, item_typ)
      call buffer_print_string(sbr, 2, ', ')
      call print_simple(i, j, k, dim1, 3, 3, item_typ)
      
      return
      end
      
      
      subroutine print_tuple(i, j, k, dim1, typ)
C     Print a tuple value of type typ.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, typ
      
C     Local variables.
      character sbr*11
      
C     External functions.
      logical pio_get_debug_mode, is_scalar_couple_type
      
      sbr = 'print_tuple'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', typ = ', typ
      endif
      
      call buffer_print_char(sbr, '(')
      
      if (is_scalar_couple_type(typ)) then
         call print_couple(i, j, k, dim1, typ)
      else
         call print_triple(i, j, k, dim1, typ)
      endif
      
      call buffer_print_char(sbr, ')')
      
      return
      end
      
      
      subroutine print_scalar(i, j, k, dim1, typ)
C     Print a scalar value.
      
      implicit none
      
C     Input arguments.
      integer i, j, k, dim1, typ
      
C     Local variable.
      character sbr*12
      
C     External functions.
      logical pio_get_debug_mode, is_scalar_simple_type
      
      sbr = 'print_scalar'
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr,
     +        ', i = ', i, ', j = ', j, ', k = ', k, ', typ = ', typ
      endif
      
      if (is_scalar_simple_type(typ)) then
         call print_simple(i, j, k, dim1, 1, 1, typ)
      else
         call print_tuple(i, j, k, dim1, typ)
      endif
      
      return
      end
      
      
      subroutine print_scalar_item(oc, i, j, k, dim1, item_typ)
C     Print a scalar item.
      
      implicit none
      
C     Input arguments.
      integer oc, i, j, k, dim1, item_typ
      
C     Local variable.
      character sbr*17
      
      sbr = 'print_scalar_item'
      
      call print_scalar(i, j, k, dim1, item_typ)
      
      call buffer_print_char(sbr, ';')
      
C      call buffer_print_newline(sbr)
C     We omit previous call to buffer_print_newline since we must flush
C     the  buffer (in order to keep it small enough), and flush_buffer
C     adds an end_of_line marker (because we use a formatted write to
C     flush the buffer).
      call flush_buffer(sbr, oc)
      
      return
      end
      
      
      subroutine print_scalar_items(oc, i, j, item_typ, sz)
C     Print scalar items of an array.
      
C     Input arguments.
      integer oc, i, j, item_typ, sz
      
C     Local variable.
      integer k
      
      do k = 1, sz
         call print_scalar_item(oc, i, j, k, sz, item_typ)
      enddo
      
      return
      end
      
      
      subroutine print_vector(oc, i, typ)
C     Print a vector of type typ.
      
      implicit none
      
C     Input arguments.
      integer oc, i, typ
      
C     Local variables.
      character sbr*12
      integer item_typ, sz(1)
      
C     External function.
      logical pio_get_debug_mode
      
      sbr = 'print_vector'
      
      item_typ = MAKE_VECTOR_ITEM_TYPE(typ)
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr, ', ',
     +        'typ = ', typ, 'item_typ = ', item_typ
      endif
      
      call buffer_print_string(sbr, 2, '[1')
      call buffer_print_newline(sbr)
      
      call fetch_value_sizes(sbr, PIO_OUT, i, sz)
      
      call buffer_print_size(sbr, sz(1))
      call buffer_print_newline(sbr)
      
      call print_scalar_items(oc, i, 1, item_typ, sz(1))
      
      call buffer_print_string(sbr, 2, '1]')
      
      return
      end
      
      
      subroutine print_multi_d_sizes(p, sizev)
C     Print the sizes of a multi dimensional array value.
      
      implicit none
      
C     Input arguments.
      integer p
      integer sizev(MAX_ARRAY_DIMENSION)
      
C     Local variable.
      character sbr*19
      
      sbr = 'print_multi_d_sizes'
      
      call buffer_print_char(sbr, '<')
      call buffer_print_natural_list2(sbr, p, sizev)
      call buffer_print_char(sbr, '>')
      
      return
      end
      
      
      subroutine print_multi_d_layout(i, p, lkv)
C     Print the layout of a multi dimensional array value.
      
      implicit none
      
C     Input arguments.
      integer i, p, lkv
      
C     Local variables.
      character sbr*20
      integer layoutv(MAX_ARRAY_DIMENSION)
      
      sbr = 'print_multi_d_layout'
      
      if (lkv .eq. C_LAYOUT_KIND) then
         
         call buffer_print_char(sbr, 'C')
         
      else if (lkv .eq. FORTRAN_LAYOUT_KIND) then
         
         call buffer_print_char(sbr, 'F')
         
      else if (lkv .eq. PERMUTATION_LAYOUT_KIND) then
         
         call fetch_value_layout(sbr, PIO_OUT, i, layoutv)
         call buffer_print_natural_list2(sbr, p, layoutv)
         
      else
         
         call error_unbound_layout_kind(sbr, lkv)
         
      endif
      
      return
      end
      
      
      subroutine print_line(oc, i, j, item_typ, sz)
C     Print the jth line (of size sz) of the ith value which is a multi
C     dimensional array value.
      
      implicit none
      
C     Input arguments.
      integer oc, i, j, item_typ, sz
      
C     Local variable.
      character sbr*10
      
      sbr = 'print_line'
      
      call buffer_print_string(sbr, 2, '[|')
      call buffer_print_newline(sbr)
      call print_scalar_items(oc, i, j, item_typ, sz)
      call buffer_print_string(sbr, 2, '|]')
      
      return
      end
      
      
      subroutine print_line_item(oc, i, j, item_typ, sz)
C     Print the jth line item (of size sz) of the ith value which is a multi
C     dimensional array value.
      
      implicit none
      
C     Input arguments.
      integer oc, i, j, item_typ, sz
      
C     Local variable.
      character sbr*15
      
      sbr = 'print_line_item'
      
      call print_line(oc, i, j, item_typ, sz)
      call buffer_print_char(sbr, ';')
      call buffer_print_newline(sbr)
      
      return
      end
      
      
      subroutine print_line_items(oc, i, item_typ, p, szs)
C     Print the lines of the ith value which is a multi dimensional
C     array value of dimension p with size szs(j) in direction j.
      
      implicit none
      
C     Input arguments.
      integer oc, i, item_typ, p, szs(MAX_ARRAY_DIMENSION)
      
C     Local variable.
      integer j
      
      do j = 1, p
         call print_line_item(oc, i, j, item_typ, szs(j))
      enddo
      
      return
      end
      
      
      subroutine print_multi_d(oc, i, typ)
C     Print ith value which is a multi dimensional array value of type
C     typ.
      
      implicit none
      
C     Input arguments.
      integer oc, i, typ
      
C     Local variables.
      character sbr*13
      integer item_typ, p, sizev(MAX_ARRAY_DIMENSION), lkv
      
C     External functions.
      logical pio_get_debug_mode
      integer fetch_value_dimension, fetch_value_layout_kind

      
      sbr = 'print_multi_d'
      
      item_typ = MAKE_MULTI_D_ITEM_TYPE(typ)
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In ', sbr, ', ',
     +        'typ = ', typ, 'item_typ = ', item_typ
      endif
      
      p = fetch_value_dimension(sbr, PIO_OUT, i)
      
      call buffer_print_char(sbr, '[')
      call buffer_print_natural(sbr, p)
      call buffer_print_newline(sbr)
      
      call fetch_value_sizes(sbr, PIO_OUT, i, sizev)
      lkv = fetch_value_layout_kind(sbr, PIO_OUT, i)
      
      call print_multi_d_sizes(p, sizev)
      call buffer_print_newline(sbr)
      call print_multi_d_layout(i, p, lkv)
      call buffer_print_newline(sbr)
      
      call print_line_items(oc, i, item_typ, p, sizev)
      
      call buffer_print_natural(sbr, p)
      call buffer_print_char(sbr, ']')
      
      return
      end
      
      
      subroutine print_array(oc, i, typ)
C     Print ith value which is an array value of type typ.
C     We already know that typ is not a scalar type.
      
      implicit none
      
C     Input arguments.
      integer oc, i, typ
      
C     External function.
      logical is_vector_type
      
      if (is_vector_type(typ)) then
         call print_vector(oc, i, typ)
      else
         call print_multi_d(oc, i, typ)
      endif
      
      return
      end
      
      
      subroutine print_array_item(oc, i, typ)
C     Print ith value which is an array value of type typ.
C     We already know that typ is not a scalar type.
      
      implicit none
      
C     Input arguments.
      integer oc, i, typ
      
C     Local variable.
      character sbr*16
      
      sbr = 'print_array_item'
      
      call print_array(oc, i, typ)
      call buffer_print_char(sbr, ';')
      call buffer_print_newline(sbr)
      
      return
      end
      
      
      subroutine print_value(oc, i, typ)
C     Print the ith value which has type typ.
      
      implicit none
      
C     Input arguments.
      integer oc, i, typ
      
C     External function.
      logical is_scalar_type
      
C     Local variable.
C      character sbr*11
      
C      sbr = 'print_value'
      
      if (is_scalar_type(typ)) then
         call print_scalar_item(oc, i, 1, 1, 1, typ)
      else
         call print_array_item(oc, i, typ)
      endif
      
      return
      end
      
      
      subroutine print_value_begin(sbr)
C     Print the "begin" marker of a typed value.
      
      implicit none
      
C     Input argument.
      character sbr*(*)
      
      call buffer_print_string(sbr, 5, 'begin')
      
      return
      end
      
      
      subroutine print_value_end(sbr)
C     Print the "end" marker of a typed value.
      
      implicit none
      
C     Input argument.
      character sbr*(*)
      
      call buffer_print_string(sbr, 3, 'end')
      
      return
      end
      
      
      subroutine print_typed_value(oc, i)
C     Print a typed value.
      
      implicit none
      
C     Input arguments.
      integer oc, i
      
C     Local variables.
      character sbr*17
      integer typ
      
C     External functions.
      logical pio_get_debug_mode
      integer fetch_value_type
      
      sbr = 'print_typed_value'
   
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'In print_typed_value, i = ', i
      endif
      
      call print_value_begin(sbr)
      call buffer_print_newline(sbr)
      
      typ = fetch_value_type(sbr, PIO_OUT, i)
      
      call print_value_type(i, typ)
      call buffer_print_newline(sbr)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'Back to print_typed_value'
      endif
      
      call print_value(oc, i, typ)
      
      if (pio_get_debug_mode()) then
         write(PIO_DEBUG, *) 'value printed'
      endif
      
      call print_value_end(sbr)
      call buffer_print_newline(sbr)
      call buffer_print_newline(sbr)
      
      return
      end
